"""Utility functions for arithmetic operations."""

__all__ = ["clip", "frange", "discretize", "counter", "ConditionChecker"]

import itertools
import math
from typing import Generator, Literal, Optional, TypeVar, overload

import numpy as np

from ..typing import QuantityValue, SupportsComparison


T = TypeVar("T", bound=SupportsComparison)


@overload
def clip(value: T, minimum: T, maximum: T) -> T:
    ...


@overload
def clip(value: T, *, absmax: T) -> T:
    ...


def clip(
    value: T,
    minimum: Optional[T] = None,
    maximum: Optional[T] = None,
    *,
    absmax: Optional[T] = None,
) -> T:
    """Limit the ``value`` to the range [``minimum``, ``maximum``].

    Parameters
    ----------
    value
        Arbitrary parameter.
    minimum
        Lower limit of ``value``.
    maximum
        Upper limit of ``value``.
    absmax
        Upper limit of absolute value of ``value``.

    Examples
    --------
    >>> neclib.utils.clip(1.2, 0, 1)
    1
    >>> neclib.utils.clip(41, 0, 100)
    41
    >>> neclib.utils.clip(-4, absmax=3)
    -3

    """
    if absmax is not None:
        maximum = abs(absmax)  # type: ignore
        minimum = -1 * maximum  # type: ignore
    if minimum > maximum:  # type: ignore
        raise ValueError("Minimum should be less than maximum.")
    return min(max(minimum, value), maximum)  # type: ignore


def frange(
    start: QuantityValue,
    stop: QuantityValue,
    step: Optional[QuantityValue] = None,
    *,
    inclusive: bool = False,
) -> Generator[QuantityValue, None, None]:
    """Float version of built-in ``range``, with support for including stop value.

    Parameters
    ----------
    start
        First value to be yielded.
    stop
        Last value to be yielded never exceeds this limit.
    step
        Difference between successive 2 values to be yielded.
    inclusive
        If ``True``, ``stop`` value can be yielded when ``stop - start`` is multiple of
        ``step``.

    Notes
    -----
    Because of floating point overflow, errors may appear when ``print``-ing the result,
    but it's the same as almost equivalent function ``numpy.arange``.

    Examples
    --------
    >>> list(neclib.utils.frange(0, 1, 0.2))
    [0, 0.2, 0.4, 0.6, 0.8]
    >>> list(neclib.utils.frange(0, 1, 0.2, inclusive=True))
    [0, 0.2, 0.4, 0.6, 0.8, 1]

    """
    if step is None:
        unity: QuantityValue = 1 * getattr(start, "unit", 1)
        step = unity
    if inclusive:
        num = -1 * np.ceil((start - stop) / step) + 1
        # HACK: ``-1 * ceil(x) + 1`` is ceiling function, but if ``x`` is integer,
        # return ``ceil(x) + 1``, so no ``x`` satisfies ``quasi_ceil(x) == x``.
    else:
        num = np.ceil((stop - start) / step)

    for i in range(int(num)):
        yield start + (step * i)


def discretize(
    value: float,
    start: float = 0.0,
    step: float = 1.0,
    *,
    method: Literal["nearest", "ceil", "floor"] = "nearest",
) -> float:
    """Convert ``value`` to nearest element of arithmetic sequence.

    Parameters
    ----------
    value
        Parameter to discretize.
    start
        First element of element of arithmetic sequence.
    step
        Difference between the consecutive 2 elements of the sequence.
    method
        Discretizing method.

    Examples
    --------
    >>> neclib.utils.discretize(3.141592)
    3
    >>> neclib.utils.discretize(3.141592, step=10)
    0
    >>> neclib.utils.discretize(3.141592, method="ceil")
    4
    >>> neclib.utils.discretize(3.141592, start=2.5, step=0.7)
    3.2

    """
    discretizer = {"nearest": round, "ceil": math.ceil, "floor": math.floor}
    return discretizer[method]((value - start) / step) * step + start


def counter(
    stop: Optional[int] = None, allow_infty: bool = False
) -> Generator[int, None, None]:
    """Generate integers from 0 to ``stop``.

    Parameters
    ----------
    stop
        Number of yielded values.
    allow_infty
        If ``True``, the counter counts up to infinity. Listing such object will cause
        memory leak, so use caution.

    Examples
    --------
    >>> list(neclib.utils.counter(5))
    [0, 1, 2, 3, 4]
    >>> list(neclib.utils.counter())
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, ...]  # -> memory leak

    """
    if (stop is None) and (not allow_infty):
        raise ValueError("Specify ``stop`` value, unless ``allow_infty`` is set True.")
    elif stop is None:
        yield from itertools.count()
    elif stop < 0:
        raise ValueError("Stop value should be non-negative.")
    else:
        yield from range(stop)


class ConditionChecker:
    def __init__(self, sequential: int = 1, reset_on_failure: bool = True):
        self.__sequential = sequential
        self.__reset_on_failure = reset_on_failure
        self.__count = 0

    def check(self, condition: bool):
        self.__count += 1 if condition else 0
        if self.__reset_on_failure:
            self.__count *= int(condition)
        return self.__count >= self.__sequential



__all__ = ["read_file"]

import os
from pathlib import Path
from typing import Union, overload
from urllib.parse import urlparse
from urllib.request import urlopen


@overload
def read_file(
    name: str, /, saveto: None, overwrite: bool, localonly: bool
) -> Union[str, bytes]:
    ...


@overload
def read_file(
    name: str, /, saveto: os.PathLike, overwrite: bool, localonly: bool
) -> None:
    ...


def read_file(name, /, saveto=None, overwrite=False, localonly=False):
    name = str(name)
    if urlparse(name).scheme:
        if localonly and (urlparse(name).scheme != "file"):
            raise FileNotFoundError(f"{name} appears to be a remote file.")
        with urlopen(name) as response:
            contents = response.read().decode("utf-8")
    elif Path(name).exists():
        path = Path(name)
        try:
            contents = path.read_text()
        except UnicodeDecodeError:
            contents = path.read_bytes()
    else:
        raise FileNotFoundError(f"Cannot find {name!r}.")

    if saveto is not None:
        save_path = Path(saveto)
        if save_path.exists() and not overwrite:
            raise FileExistsError(f"{save_path} already exists.")

        save_path.parent.mkdir(parents=True, exist_ok=True)
        save_path.touch(exist_ok=True)
        try:
            save_path.write_text(contents)
        except TypeError:
            save_path.write_bytes(contents)
        return
    return contents



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['neclib',
 'neclib.controllers',
 'neclib.coordinates',
 'neclib.data',
 'neclib.devices',
 'neclib.devices.attenuator',
 'neclib.devices.bias_reader',
 'neclib.devices.bias_setter',
 'neclib.devices.encoder',
 'neclib.devices.motor',
 'neclib.devices.power_meter',
 'neclib.devices.signal_generator',
 'neclib.devices.spectrometer',
 'neclib.devices.thermometer',
 'neclib.devices.vacuum_gauge',
 'neclib.devices.weather_station',
 'neclib.interfaces',
 'neclib.parameters',
 'neclib.parameters.parser',
 'neclib.recorders',
 'neclib.safety',
 'neclib.simulators',
 'neclib.utils']

package_data = \
{'': ['*'], 'neclib': ['src/*']}

install_requires = \
['astropy>=5.0.4,<6.0.0',
 'necstdb>=0.2.8,<0.3.0',
 'numpy>=1.22,<2.0',
 'ogameasure>=0.5.4,<0.6.0',
 'tomlkit>=0.11.4,<0.12.0',
 'xfftspy>=0.1.3,<0.2.0']

extras_require = \
{':sys_platform == "linux"': ['pyinterface>=1.6,<2.0']}

setup_kwargs = {
    'name': 'neclib',
    'version': '0.16.1',
    'description': 'Pure Python tools for NECST.',
    'long_description': '# neclib\n\n[![PyPI](https://img.shields.io/pypi/v/neclib.svg?label=PyPI&style=flat-square)](https://pypi.org/pypi/neclib/)\n[![Python versions](https://img.shields.io/pypi/pyversions/neclib.svg?label=Python&color=yellow&style=flat-square)](https://pypi.org/pypi/neclib/)\n[![Test status](https://img.shields.io/github/actions/workflow/status/necst-telescope/neclib/test.yml?branch=main&logo=github&label=Test&style=flat-square)](https://github.com/necst-telescope/neclib/actions)\n[![codecov](https://codecov.io/gh/necst-telescope/neclib/branch/main/graph/badge.svg?token=DP2ZTYBOTR)](https://codecov.io/github/necst-telescope/neclib)\n[![License: MIT](https://img.shields.io/badge/license-MIT-blue.svg?label=License&style=flat-square)](https://github.com/necst-telescope/neclib/blob/main/LICENSE)\n\nPure Python tools for NECST.\n\n## Features\n\nThis library provides:\n\n- Miscellaneous tools for NECST system.\n\n## Installation\n\n```shell\npip install neclib\n```\n\n## Usage\n\nSee the [API Reference](https://necst-telescope.github.io/neclib/_source/neclib.html).\n\n---\n\nThis library is using [Semantic Versioning](https://semver.org).\n',
    'author': 'KaoruNishikawa',
    'author_email': 'k.nishikawa@a.phys.nagoya-u.ac.jp',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://necst-telescope.github.io/neclib/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



__all__ = ["ND287"]

from typing import Literal

import astropy.units as u
import ogameasure

from ... import utils
from .encoder_base import Encoder


class ND287(Encoder):
    """Encoder readout."""

    Manufacturer = "HEIDENHAIN"
    Model = "ND287"

    Identifier = "port_az"

    @utils.skip_on_simulator
    def __init__(self) -> None:
        self.driver = {
            "az": ogameasure.HEIDENHAIN.ND287(self.Config.port_az),
            "el": ogameasure.HEIDENHAIN.ND287(self.Config.port_el),
        }

    def get_reading(self, axis: Literal["az", "el"]) -> u.Quantity:
        raw = self.driver[axis.lower()].output_position_display_value()
        return float(raw.strip(b"\x02\x00\r\n").decode()) << u.deg  # type: ignore

    def finalize(self) -> None:
        pass



from abc import abstractmethod

import astropy.units as u

from ..device_base import DeviceBase


class VacuumGauge(DeviceBase):
    @abstractmethod
    def get_pressure(self) -> u.Quantity:
        ...



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, Type


class Writer(ABC):

    _instance: Dict[Type["Writer"], "Writer"] = {}
    _initialized: Dict[Type["Writer"], bool] = {}

    def __new__(cls, *args, **kwargs):
        if cls._instance.get(cls) is None:
            cls._instance[cls] = super().__new__(cls)
            cls._initialized[cls] = False
        return cls._instance[cls]

    @abstractmethod
    def start_recording(self, record_dir: Path) -> None:
        ...

    @abstractmethod
    def append(self, *args: Any, **kwargs: Any) -> bool:
        """Append a chunk of data.

        Returns
        -------
        handled
            Whether the data is handled by this writer.

        Notes
        -----
        The subclass should check the input type, as ``Writer`` would handle some
        essentially different data, and ``Recorder`` won't implement the type checking
        for extensibility.

        """
        ...

    @abstractmethod
    def stop_recording(self) -> None:
        ...



"""Pure Python tools for NECST."""


# Perform Time-consuming Downloads
class _TimeConsumingTasks:
    @staticmethod
    def download_astropy_parameter_files():
        """Will download finals2000A.all (3.4MB) and Leap_Second.dat (1.3KB)."""
        from astropy.time import Time

        Time.now().ut1


import concurrent.futures  # noqa: E402

executor = concurrent.futures.ThreadPoolExecutor()
futures = [
    executor.submit(getattr(_TimeConsumingTasks, func))
    for func in dir(_TimeConsumingTasks)
    if not func.startswith("_")
]


# Logger configuration
import logging  # noqa: E402

rootLogger = logging.getLogger()
rootLogger.setLevel(logging.DEBUG)
del logging, rootLogger


# Version Definition
from importlib.metadata import version  # noqa: E402

try:
    __version__ = version("neclib")
except:  # noqa: E722
    __version__ = "0.0.0"  # Fallback.
del version


# Environment Variables
class EnvVarName:
    necst_root: str = "NECST_ROOT"
    necst_config_name: str = "NECST_CONFIG_NAME"
    ros2_ws: str = "ROS2_WS"
    domain_id: str = "ROS_DOMAIN_ID"
    record_root: str = "NECST_RECORD_ROOT"


# Warn Restriction Imposed by Environment
import sys  # noqa: E402

from .interfaces import get_logger  # noqa: E402

logger = get_logger("neclib")
if sys.platform != "linux":
    logger.warning(
        "Device drivers for Interface PCI boards are only supported on Linux."
    )
del logger, sys  # get_logger is intentionally kept in the namespace.


# Subpackages
from . import controllers  # noqa: F401, E402
from . import interfaces  # noqa: F401, E402
from . import parameters  # noqa: F401, E402
from . import recorders  # noqa: F401, E402
from . import safety  # noqa: F401, E402
from . import simulators  # noqa: F401, E402
from . import typing  # noqa: F401, E402
from . import units  # noqa: F401, E402
from . import utils  # noqa: F401, E402

# Read Configuration
from .configuration import config, configure  # noqa: F401, E402

# Aliases
from .exceptions import *  # noqa: F401, E402, F403

# Wait for all background tasks to complete.
concurrent.futures.wait(futures, timeout=60)
executor.shutdown()
del _TimeConsumingTasks, concurrent, executor, futures


# Make exception informative
import threading  # noqa: E402


def format_exc(args) -> None:
    import traceback

    tb = "\n".join(traceback.format_tb(args.exc_traceback, 5))
    get_logger(__name__).error(
        f"Unhandled exception in thread {getattr(args.thread, 'name', 'unknown')}:\n"
        f"\033[1mType:\033[0m {args.exc_type.__name__}\n"
        f"\033[1mValue:\033[0m {args.exc_value}\n"
        f"\033[1mTraceback:\033[0m\n{tb}",
    )


threading.excepthook = format_exc
del threading, format_exc



"""Storage for pointing error parameter."""

__all__ = ["PointingErrorData"]

from pathlib import Path
from typing import Any, Dict, List, Tuple

import astropy.units as u
import tomlkit
from astropy.coordinates import Angle

from ...typing import PathLike
from ...utils import ParameterMapping, read_file


class PointingErrorData(ParameterMapping):
    """Parse pointing error parameter as Quantity.

    Attributes
    ----------
    dAz: Angle
        Azimuth (not X) offset of encoder reading.
    de: Angle
        X collimation error.
    chi_Az: Angle
        Magnitude of tilt of azimuth axis.
    omega_Az: Angle
        Phase (azimuth direction) of tilt of azimuth axis.
    eps: Angle
        Skew angle (lack of orthogonality) between azimuth and elevation axes.
    chi2_Az: Angle
        Same as chi, but the period is 180deg in azimuth axis (harmonic component).
    omega2_Az: Angle
        Same as omega, but the period is 180deg in azimuth (harmonic component).
    chi_El: Angle
        Amplitude of tilt of azimuth axis (same as chi_Az).
    omega_El: Angle
        Phase (azimuth direction) of tilt of azimuth axis (same as omega_Az).
    chi2_El: Angle
        Same as chi, but the period is 180deg in azimuth axis (harmonic component,
        same as chi2_Az).
    omega2_El: Angle
        Same as omega, but the period is 180deg in azimuth (harmonic component, same
        as omega2_Az).
    g: float
        First order gravitational deflection coefficient.
    gg: float
        Second order gravitational deflection coefficient.
    ggg: float
        Third order gravitational deflection coefficient.
    gggg: float
        Fourth order gravitational deflection coefficient.
    dEl: Angle
        Elevation offset of encoder reading.
    de_radio: Angle
        Constant X (not azimuth) offset between optical and radio beam.
    del_radio: Angle
        Constant elevation offset between optical and radio beam.
    cor_v: Angle
        Amplitude of collimation error.
    cor_p: Angle
        Phase of collimation error, negative of elevation where the elevation
        component of collimation error is zero.
    g_radio: float
        First order gravitational deflection coefficient.
    gg_radio: float
        Second order gravitational deflection coefficient.
    ggg_radio: float
        Third order gravitational deflection coefficient.
    gggg_radio: float
        Fourth order gravitational deflection coefficient.

    """

    def __init__(self, **kwargs: Dict[str, Any]) -> None:
        kwargs = self._make_quantity(kwargs)
        super().__init__(**kwargs)

    @classmethod
    def from_file(cls, path: PathLike):
        """Parse TOML file.

        Parameters
        ----------
        path
            Path to parameter file.

        Notes
        -----
        Bare keys or nested tables are not allowed. Valid parameters are those declared
        in the following format. The table structure will be flattened, so the
        ``parameter kind`` won't be preserved.

        .. code-block:: toml

           [parameter kind]
           name = value

        Examples
        --------
        >>> params = neclib.parameters.PointingErrorData.from_file("path/to/error_param.toml")
        >>> params
        ObsParams({'dAz': <Quantity 5300. arcsec>, 'de': <Quantity 380. arcsec>, ...})
        >>> params.dAz
        <Quantity 5300. arcsec>
        >>> params["de"]
        <Quantity 380. arcsec>

        """  # noqa: E501
        _params = tomlkit.parse(read_file(path))
        params = {}
        _ = [params.update(subdict) for subdict in _params.values()]
        return cls(**params)

    @classmethod
    def from_text_file(cls, path: PathLike):
        """Parse legacy style pointing error parameter file.

        Parameters
        ----------
        path
            Path to parameter file.

        Examples
        --------
        >>> params = neclib.parameters.PointingErrorData.from_text_file("path/to/error_param.txt")
        >>> params
        ObsParams({'dAz': <Quantity 5300. arcsec>, 'de': <Quantity 380. arcsec>, ...})
        >>> params.dAz
        <Quantity 5300. arcsec>
        >>> params["de"]
        <Quantity 380. arcsec>

        """  # noqa: E501
        lines = Path(path).read_text().split("\n")
        filled_lines = [line for line in lines if line != ""]
        parameter_units: List[Tuple[str, str]] = [
            ("dAz", "arcsec"),
            ("de", "arcsec"),
            ("chi_Az", "arcsec"),
            ("omega_Az", "deg"),
            ("eps", "arcsec"),
            ("chi2_Az", "arcsec"),
            ("omega2_Az", "deg"),
            ("chi_El", "arcsec"),
            ("omega_El", "deg"),
            ("chi2_El", "arcsec"),
            ("omega2_El", "deg"),
            ("g", ""),
            ("gg", ""),
            ("ggg", ""),
            ("gggg", ""),
            ("dEl", "arcsec"),
            ("de_radio", "arcsec"),
            ("del_radio", "arcsec"),
            ("cor_v", "arcsec"),
            ("cor_p", "deg"),
            ("g_radio", ""),
            ("gg_radio", ""),
            ("ggg_radio", ""),
            ("gggg_radio", ""),
        ]
        parameters = {
            name: value + unit
            for value, (name, unit) in zip(filled_lines, parameter_units)
        }
        return cls(**parameters)

    @staticmethod
    def _make_quantity(parameters: Dict[str, Any]) -> Dict[str, Any]:
        parsed = {}
        for name, value in parameters.items():
            if value == {}:  # Empty value
                pass
            else:
                try:
                    parsed[name] = Angle(value)
                except (u.UnitsError, u.UnitTypeError):
                    parsed[name] = float(value)
        return parsed



from typing import Literal

import astropy.units as u
import ogameasure

from .attenuator_base import Attenuator


class RHIO10(Attenuator):

    Manufacturer: str = "SENA"
    Model: str = "RHIO10"

    Identifier = "host"

    def __init__(self) -> None:
        com = ogameasure.ethernet(self.Config.host, self.Config.port)
        self.io = ogameasure.SENA.adios(com)

    def get_level(self, ch: Literal[1, 2]) -> u.Quantity:
        if ch == 1:
            return self.io.get_att1() << u.dB
        elif ch == 2:
            return self.io.get_att2() << u.dB
        raise ValueError(f"Invalid channel number: {ch}")

    def set_level(self, level_dB: int, ch: Literal[1, 2]) -> None:
        self.io._set_att(ch, int(level_dB))

    def finalize(self) -> None:
        self.io.com.close()



import queue
import time
import traceback
from pathlib import Path
from threading import Event, Thread
from typing import Any, Callable, Dict, Iterable, List, Optional, Sequence, Tuple, Union

import necstdb

from .. import get_logger
from ..typing import TextLike
from .writer_base import Writer

TextData = Union[TextLike, List[TextLike]]


def parse_str_size(data: TextData) -> Tuple[str, int]:
    length = len(data) if isinstance(data, TextLike) else max(map(len, data))
    return f"{length}s", 1 * length


def str_to_bytes(data: TextData) -> Union[bytes, List[bytes]]:
    if isinstance(data, TextLike):
        return data if isinstance(data, bytes) else data.encode("utf-8")  # type: ignore
    else:
        return [str_to_bytes(elem) for elem in data]  # type: ignore


class NECSTDBWriter(Writer):
    """Dump data to NECSTDB.

    Attributes
    ----------
    db : necstdb.necstdb.necstdb or None
        Database instance this writer is currently dumping data to.
    tables : Dict[str, necstdb.necstdb.table]
        NECSTDB Table to which this writer is dumping data.

    Examples
    --------
    >>> writer = neclib.recorders.NECSTDBWriter()
    >>> writer.start_recording("path/to/database/directory.necstdb")
    >>> chunk = [
    ...     {"key": "timestamp", "type": "double", "value": 1664195057.022712},
    ...     {"key": "pressure", "type": "float32", "value": 850.5},
    ... ]
    >>> writer.append("/observatory/pressure", chunk)
    >>> writer.stop_recording()

    """

    LivelinessDuration: float = 15.0
    """If a table isn't updated for this duration (in sec), it will be closed."""

    WarningQueueSize: int = 1000
    """Warn if number of data waiting for being dumped is greater than this."""

    DTypeConverters: Dict[str, Callable[[Any], Tuple[Any, str, int]]] = {
        "bool": lambda dat: (dat, "?", 1),
        "byte": lambda dat: (dat, *parse_str_size(dat)),
        "char": lambda dat: (dat, "c", 1),
        "float32": lambda dat: (dat, "f", 4),
        "float": lambda dat: (dat, "f", 4),
        "float64": lambda dat: (dat, "d", 8),
        "double": lambda dat: (dat, "d", 8),
        "int8": lambda dat: (dat, "b", 1),
        "int16": lambda dat: (dat, "h", 2),
        "int32": lambda dat: (dat, "i", 4),
        "int64": lambda dat: (dat, "q", 8),
        "uint8": lambda dat: (dat, "B", 1),
        "uint16": lambda dat: (dat, "H", 2),
        "uint32": lambda dat: (dat, "I", 4),
        "uint64": lambda dat: (dat, "Q", 8),
        "string": lambda dat: (str_to_bytes(dat), *parse_str_size(dat)),
    }
    """Converter from readable type name to C data structure."""

    def __init__(self) -> None:
        if not self._initialized[self.__class__]:
            self.logger = get_logger(self.__class__.__name__)

            self.db: Optional[necstdb.necstdb.necstdb] = None
            self.tables: Dict[str, necstdb.necstdb.table] = {}

            self.recording_path: Optional[Path] = None

            self._data_queue = queue.Queue()
            self._thread = None

            self._stop_event: Optional[Event] = None
            self._table_last_update: Dict[str, float] = {}

            self._initialized[self.__class__] = True

    def start_recording(self, record_dir: Path) -> None:
        self.recording_path = record_dir

        # Use root directory of current record as database.
        self.db = necstdb.opendb(record_dir, mode="w")

        self._stop_event = Event()
        self._thread = Thread(target=self._update_background, daemon=True)
        self._thread.start()

    def append(
        self,
        topic: Optional[str] = None,
        chunk: Optional[List[Dict[str, Any]]] = None,
        *args,
        **kwargs,
    ) -> bool:
        """Append a chunk of data.

        Returns
        -------
        handled
            Whether the data is handled by this writer.

        Examples
        --------
        >>> chunk = [
                {"key": "timestamp", "type": "double", "value": 1664195057.022712},
                {"key": "reading", "type": "int32", "value": 5},
            ]
        >>> writer.append("/meter_reading", chunk)

        """
        if not (
            isinstance(chunk, Iterable)
            and all([isinstance(field, dict) for field in chunk])
            and all([{"key", "type", "value"} <= field.keys() for field in chunk])
        ):  # This writer only handles data passes this condition.
            return False
        if self.db is None:
            self.logger.warning("Database is not opened. Data will be lost.")
            return False

        self._data_queue.put((topic, chunk))
        return True

    def stop_recording(self) -> None:
        if self._stop_event is not None:
            self._stop_event.set()
        if self._thread is not None:
            self._thread.join()

        self._stop_event = self._thread = None

        self.db = None
        [table.close() for table in self.tables.values()]
        self.tables.clear()
        self._table_last_update.clear()
        self.recording_path = None

    def _update_background(self) -> None:
        while (self._stop_event is not None) and (not self._stop_event.is_set()):
            self._check_liveliness()
            if self._data_queue.empty():
                time.sleep(0.01)
                continue
            if self._data_queue.qsize() > self.WarningQueueSize:
                self.logger.warning("Too many data waiting for being dumped.")
            topic, chunk = self._data_queue.get()
            self._write(topic, chunk)

        if not self._data_queue.empty():
            qsize = self._data_queue.qsize()
            self.logger.info(f"Dumping received data: {qsize} remaining...")
        while not self._data_queue.empty():
            topic, chunk = self._data_queue.get()
            self._write(topic, chunk)
        self.logger.info("NECSTDB has gracefully been stopped.")

    def _check_liveliness(self) -> None:
        table_names = list(self.tables.keys())
        # Conversion to list is workaround for deepcopy (cannot deepcopy keys obj)
        # Object copying avoids dict size change during iteration in the following lines

        now = time.time()
        for topic in table_names:
            if now - self._table_last_update[topic] > self.LivelinessDuration:
                self.remove_table(topic)

    def _write(self, topic: str, chunk: List[Dict[str, Any]]) -> None:
        try:
            now = time.time()
            self._table_last_update[topic] = now

            data = [now]
            metadata = [{"key": "recorded_time", "format": "d", "size": 8}]

            for field in chunk:
                parsed = self._parse_field(field)
                if parsed is None:
                    return
                _data, _metadata = parsed
                data.extend(_data)
                metadata.append(_metadata)
            if topic not in self.tables:
                record_writer = self.__class__.__name__
                self.add_table(
                    topic,
                    {"data": metadata, "memo": f"Generated by {record_writer}"},
                )
            self.tables[topic].append(*data)
        except Exception:
            self.logger.error(traceback.format_exc())

    def _parse_field(
        self, field: Dict[str, Any]
    ) -> Optional[Tuple[List[Any], Dict[str, str]]]:
        data = fmt = size = None
        for k in self.DTypeConverters:
            if field["type"].find(k) != -1:
                data, fmt, size = self.DTypeConverters[k](field["value"])
                break
        if (data is None) or (fmt is None) or (size is None):
            return

        if isinstance(data, Sequence) and (not isinstance(data, TextLike)):
            length = len(data)
            if fmt.find("s") != -1:
                fmt = "0s" if length == 0 else fmt * length
                # Because 5s5s5s is ok, but 35s has different meaning.
            else:
                fmt = f"{length}{fmt}"
            size *= length
            data = list(data)
        else:
            data = [data]
        return data, {"key": field["key"], "format": fmt, "size": size}

    def _validate_name(self, topic: str) -> str:
        return topic.replace("/", "-").strip("-")

    def add_table(self, topic: str, metadata: Dict[str, Any]) -> None:
        if self.db is None:
            raise RuntimeError("Database is not opened.")

        _topic = self._validate_name(topic)
        if topic in self.tables:
            self.logger.warning(f"Table {topic} already opened.")
            return
        if _topic not in self.db.list_tables():
            metadata.update({"necstdb_version": necstdb.__version__})
            self.db.create_table(_topic, metadata)
        self.tables[topic] = self.db.open_table(_topic, mode="ab")
        self._table_last_update[topic] = time.time()

    def remove_table(self, topic: str) -> None:
        if topic in self.tables:
            self.tables[topic].close()
        self.tables.pop(topic, None)
        self._table_last_update.pop(topic, None)



"""Storage for observation parameter."""

__all__ = ["ObsParamData"]

from typing import Any, Dict

import tomlkit
from astropy.coordinates import Angle
from astropy.units import Quantity

from ...typing import PathLike
from ...utils import ParameterMapping, read_file


class ObsParamData(ParameterMapping):
    """Parse observation spec as Quantity.

    Parameters named uppercase (``A-Z0-9_``) will be parsed as it is, lowercase
    (``a-z0-9_``) will be parsed as ``Quantity``, mixed case (``[A-Z]+[a-z]+0-9_``) will
    be parsed as ``Angle``.

    """

    def __init__(self, **kwargs: Dict[str, Any]) -> None:
        kwargs = self._make_quantity(kwargs)
        super().__init__(**kwargs)

    @classmethod
    def from_file(cls, path: PathLike):
        """Parse TOML file.

        Parameters
        ----------
        path
            Path to parameter file.

        Notes
        -----
        Bare keys or nested tables are not allowed. Valid parameters are those declared
        in the following format. The table structure will be flattened, so the
        ``parameter kind`` won't be preserved.

        .. code-block:: toml

           [parameter kind]
           name = value

        Examples
        --------
        >>> params = neclib.parameters.ObsParamData.from_file("path/to/spec.obs.toml")
        >>> params
        ObsParams({'OBSERVER': 'amigos', 'OBJECT': 'OriKL', ...})
        >>> params.OBSERVER
        'amigos'
        >>> params["OBJECT"]
        'OriKL'

        """
        _params = tomlkit.parse(read_file(path))
        params = {}
        _ = [params.update(subdict) for subdict in _params.values()]
        return cls(**params)

    @staticmethod
    def _make_quantity(parameters: Dict[str, Any]) -> Dict[str, Any]:
        parsed = {}
        for name, value in parameters.items():
            if value == {}:  # Empty value
                pass
            elif name.isupper():
                parsed[name] = value
            elif name.islower():
                parsed[name] = Quantity(value)
            else:
                parsed[name] = Angle(value)
        return parsed
